/*----------------------------------------------------------------------
This source distribution is placed in the public domain by its author,
Ben Buhrow. You may use it for any purpose, free of charge,
without having to notify anyone. I disclaim any responsibility for any
errors.

Optionally, please be nice and tell me if you find this source to be
useful. Again optionally, if you add to the functionality present here
please consider making those additions public too, so that others may 
benefit from your work.
----------------------------------------------------------------------*/


//symbols in calc
#define EOE 1
#define IMM 2
#define NUM 3
#define OP 4
#define RP 5
#define LP 6
#define CH 7
#define AMBIG 8
#define COMMA 9
#define SPACE 10

//operator associativity
#define RIGHT 1
#define LEFT 0

function get_el_type2(char s) {
	//there are several types of characters in an expression.	
	//decide which type this is
	if (isdigit(s) || (s <= 90 && s >= 65)) return NUM;
	else if (s == '(') return LP;
	else if (s == ')') return RP;
	else if (s == '-') return AMBIG;
	else if (getIMM(s) >= 0) return IMM;
	else if (getOP(s) >= 0) return OP;
	else if (isEOE(s)) return EOE;
	else if (s == ',') return COMMA;
	else if ((s <= 122 && s >= 95) || s == 39) return CH;
	else if (isspace(s)) return SPACE;
	else return -1;
}

function isEOE(char s) {
	if (s == 0) return 1;
	else return 0;
}

function getIMM(char s) {
	for (var i=0;i<3;i++) {
		if (s == imms[i]) return i;
	}
	return -1;
}

function getOP(char s) {
	//return >=0 if this char is a opchar
	for (var i=0;i<numopchars;i++) {
		if (opchar[i] == s) return i;
	}
	return -1;
}

function getAssoc(char *s) {
	if (strcmp(s,"^") == 0) return RIGHT;
	else return LEFT;
}


function op_precedence(char *s1, char *s2, int assoc) {
	//if associativity is RIGHT, then use strictly greater than
	//else use greater than or equal to
	int p1=0,p2=0;

	if (strcmp(s1,"=") == 0) p1=-1;
	if (strcmp(s1,"<<") == 0) p1=0;
	if (strcmp(s1,">>") == 0) p1=0;
	if (strcmp(s1,"+") == 0) p1=1;
	if (strcmp(s1,"-") == 0) p1=1;
	if (strcmp(s1,"*") == 0) p1=2;
	if (strcmp(s1,"/") == 0) p1=2;
	if (strcmp(s1,"%") == 0) p1=2;
	if (strcmp(s1,"^") == 0) p1=3;
	if (strcmp(s1,"\\") == 0) p1=4;

	if (strcmp(s1,"=") == 0) p1=-1;
	if (strcmp(s2,"<<") == 0) p2=0;
	if (strcmp(s2,">>") == 0) p2=0;
	if (strcmp(s2,"+") == 0) p2=1;
	if (strcmp(s2,"-") == 0) p2=1;
	if (strcmp(s2,"*") == 0) p2=2;
	if (strcmp(s2,"/") == 0) p2=2;
	if (strcmp(s2,"%") == 0) p2=2;
	if (strcmp(s2,"^") == 0) p2=3;
	if (strcmp(s2,"\\") == 0) p2=4;

	if (assoc == LEFT) return p1 >= p2;
	else return p1 > p2;
}

function preprocess(str_t *in) {
	//preprocess the expression in 'in'

	var j = 0;
	//remove white space
	for (var i=0;i<in->nchars;i++) {
		if (isspace(in->s[i])) continue;
		
		in->s[j] = in->s[i];
		j++;
	}
	in->s[j] = '\0';

	//look for a^b%c combinations, and replace with a function call
	//or (a^b)%c
	//should ignore stuff inside parens, so for instance
	//(3+5)^(8*9*7)%(134513%46) should reduce to
	//modexp(3+5,8*9*7,134513%46)

	//algebraic simplification (this would be cool...)

	return 0;
}

function is_new_token(int el_type, int el_type2) {

	if (el_type == EOE || el_type == LP || el_type == RP) return 1;

	if (el_type != el_type2) {
		//types are different
		if (el_type == CH && el_type2 == NUM) {
			//but this could be a function or variable name
			//so not different
			return 0;
		}
		else if (el_type == NUM && el_type2 == CH) {
			//but this could be a function or variable name
			//so not different
			return 0;
		}
		else return 1;
	}
	return 0;
}

function tokenize(char *in, int *token_types, int *num_tokens) {
	//take a string as input
	//break it into tokens
	//create an array of strings for each token
	//return the pointer to the array and the number of elements
	//in the array.	this will all have to be freed later
	//by the caller

	//a token in this context is one of the following things:
	//	a number, possibly including a base prefix (0x, 0d, 0b, etc)
	//	a variable name
	//	a function name
	//	an operator string (includes parens, commas)

	//read the string one character at a time
	//for each character read, decide if we've found the start of a new token

	int inpos, i, el_type, el_type2, token_alloc, tmpsize = GSTR_MAXSIZE;
	char ch;
	char *tmp;
	char **tokens;

	token_alloc = 100;		//100 tokens
	tokens = (char **)malloc(token_alloc * sizeof(char *));
	*num_tokens = 0;

	tmp = (char *)malloc(GSTR_MAXSIZE * sizeof(char));

	//get the first character and check the type
	inpos = 0;
	i=1;
	ch = in[inpos];
	tmp[i-1] = ch;
	el_type = get_el_type2(ch);

	//when an expression gets cast into postfix, it aquires a leading
	//space which we can skip here
	if (el_type == SPACE) {
		inpos = 1;
		i=1;
		ch = in[inpos];
		tmp[i-1] = ch;
		el_type = get_el_type2(ch);
	}

	//ambiguous types:
	//a "-" can be either a num (if a negative sign) or an operator
	//a number can be a number or a string (num or func/var name)
	//a letter can be a string or a number (hex or func/var name)
	//we can tell them apart from the surrounding context
	//
	//negative signs never have a num type before (or anything that
	//can be evaluated as a num , i.e. ")"
	//
	//if we are reading CH's don't stop interpreting them as CH's until
	//we find a non-CH or non-NUM (use a flag)
	//
	//watch for magic combinations '0x' '0d', etc.	set a flag to 
	//interpret what follows as the appropriate kind of num, and
	//discard the '0x', etc.
	//once this is fixed, change the final stack evaluation to print hex
	//strings to save some conversion time.
	if (el_type == AMBIG) {
		if (get_el_type2(in[inpos+1]) == NUM) el_type = NUM;
		else el_type = OP;
	}
	while (1) {
		//get another character and check the type
		inpos++;
		//if el_type == EOE, then no reason to keep reading.	This bug didn't seem to cause
		//any crashes, but couldn't have been healthy...
		if (el_type == EOE) break;

		ch = in[inpos];
		el_type2 = get_el_type2(ch);
		if (el_type2 == AMBIG) {
			switch (get_el_type2(in[inpos-1]))
			{
			case OP:
				el_type2 = NUM;
				break;
			case LP:
				el_type2 = NUM;
				break;
			case RP:
				el_type2 = OP;
				break;
			case CH:
				el_type2 = OP;
				break;
			case IMM:
				el_type2 = OP;
				break;
			case NUM:
				el_type2 = OP;
				break;
			case COMMA:
				el_type2 = NUM;
				break;
			case SPACE:
				//when processing postfix strings, we need this
				el_type2 = OP;
				break;
			default:
				printf("misplaced - sign\n");
				for (i=0;i< *num_tokens; i++)
					free(tokens[i]);
				free(tokens);
				free(tmp);
				return NULL;
			}
		}
		
		if (is_new_token(el_type,el_type2) || el_type == EOE) {
			if (el_type == EOE) break;

			if (el_type == -1) {
				//unrecognized character.	clear all tokens and return;
				printf("unrecognized character in input\n");
				for (i=0;i< *num_tokens; i++)
					free(tokens[i]);
				free(tokens);
				free(tmp);
				return NULL;
			}

			if (el_type != SPACE) {
				//create a new token
				tmp[i] = '\0';
				tokens[*num_tokens] = (char *)malloc((strlen(tmp) + 2) * sizeof(char));
				strcpy(tokens[*num_tokens],tmp);
				token_types[*num_tokens] = el_type;
				*num_tokens = *num_tokens + 1;

				if (*num_tokens >= token_alloc) {
					tokens = (char **)realloc(tokens, token_alloc * 2 * sizeof(char *));
					token_types = (int *)realloc(token_types, token_alloc * 2 * sizeof(int));
					token_alloc *= 2;
				}
			}
		
			//then cycle the types
			el_type = el_type2;
			i=1;
			strcpy(tmp,&ch);
		}
		else {
			if (i == (tmpsize - 1)) {
				//printf("growing tmpsize in tokenize...\n");
				tmpsize += GSTR_MAXSIZE;
				tmp = (char *)realloc(tmp,tmpsize * sizeof(char));
			}
			tmp[i] = ch;
			i++;
		}
	}

	return tokens;
}

function isNumber(char *str) {
	int base = 10;
	int first_char_is_zero = 0;

	for (var i = 0; i < strlen(str); i++) {
		if ((i == 0) && (str[i] == '0')) {
			first_char_is_zero = 1;
			continue;
		}
		if ((i == 1) && first_char_is_zero) {
			if (str[i] == 'b') base = 2;
			else if (str[i] == 'o') base = 8;
			else if (str[i] == 'd' || (str[i] >= '0' && str[i] <= '9')) base = 10;
			else if (str[i] == 'x') base = 16;
			else return 0;
			continue;
		}
		if ((base == 2) && !(str[i] >= '0' && str[i] <= '1')) return 0;
		if ((base == 8) && !(str[i] >= '0' && str[i] <= '7')) return 0;
		if ((base == 10) && !(str[i] >= '0' && str[i] <= '9')) return 0;
		if ((base == 16) && !(isdigit(str[i]) || (str[i] >= 'a' && str[i] <= 'f') || (str[i] >= 'A' && str[i] <= 'F'))) return 0;
	}
	return 1;
}

function isOperator(char *str) {
	if (str[0] == '+' || str[0] == '-' || str[0] == '*' || str[0] == '/' || str[0] == '%' || str[0] == '^') return 1;
	if (str[0] == '!' || str[0] == '#') return 1;
	return 0;
}

/* check to see if a string contains an operator */
function hasOperator(char *str) {
	for (var i = 0; i < strlen(str); i++)
		if (isOperator(&str[i]))
			return 1;
	return 0;
}

/* check to see if str contains + or - (Addition or Subtraction)*/
function hasOperatorAS(char *str) {
	for (var i = 0; i < strlen(str); i++)
		if (str[i] == '+' || str[i] == '-')
			return 1;
	return 0;
}

function get_expression(char *in, str_t *out) {
	char *tok;
	char delim[] = {' ', '\0'};
	int invalid_string = 0;
	bstack_t stk;
	str_t *tmp,*tmp1,*tmp2;

	/* if there is no input to process, we are done... */
	if (in == NULL) return;

	stack_init(20,&stk,STACK);
	tmp = (str_t *)malloc(sizeof(str_t));
	sInit(tmp);
	tmp1 = (str_t *)malloc(sizeof(str_t));
	sInit(tmp1);
	tmp2 = (str_t *)malloc(sizeof(str_t));
	sInit(tmp2);

	tok = strtok(in, delim);
	while (tok != NULL)
	{
		/* if token is a number, push it onto the stack... */
		if (isNumber(tok))
		{
			toStr(tok, tmp);
			push(tmp, &stk);
		}
		else if (isOperator(tok))
		{
			if (tok[0] == '!' || tok[0] == '#')
			{
				/* factorial and primorial are unary operators */
				if (pop(tmp1, &stk) == 0)
				{
					invalid_string = 1;
					break;
				}
				if (hasOperator(tmp1->s))
				{
					sClear(tmp);
					sAppend("(", tmp);
					sAppend(tmp1->s, tmp);
					sAppend(")", tmp);
					sAppend(tok, tmp);
				}
				else
				{
					sClear(tmp);
					sAppend(tmp1->s, tmp);
					sAppend(tok, tmp);
				}
				push(tmp, &stk);
			}
			else if (tok[0] == '+')
			{
				/* pop off two elements a,b
					 push back on "(aOPb)" */
				if ((pop(tmp2, &stk) == 0) || (pop(tmp1, &stk) == 0))
				{
					invalid_string = 1;
					break; /* bad input string, don't parse any more... */
				}
				sClear(tmp);
				sAppend(tmp1->s, tmp);
				sAppend(tok, tmp);
				sAppend(tmp2->s, tmp);
				push(tmp, &stk);
			}
			else if (tok[0] == '-')
			{
				/* pop off two elements a,b
					 push back on "(aOPb)" */
				if ((pop(tmp2, &stk) == 0) || (pop(tmp1, &stk) == 0))
				{
					invalid_string = 1;
					break; /* bad input string, don't parse any more... */
				}
				sClear(tmp);
				sAppend(tmp1->s, tmp);
				sAppend(tok, tmp);
				if (hasOperatorAS(tmp2->s))
				{
					sAppend("(", tmp);
					sAppend(tmp2->s, tmp);
					sAppend(")", tmp);
				}
				else
				{
					sAppend(tmp2->s, tmp);
				}
				push(tmp, &stk);
			}
			else
			{
				/* pop off two elements a,b
					 push back on "(aOPb)" */
				if ((pop(tmp2, &stk) == 0) || (pop(tmp1, &stk) == 0))
				{
					invalid_string = 1;
					break; /* bad input string, don't parse any more... */
				}
				sClear(tmp);
				if (hasOperator(tmp1->s))
				{
					sAppend("(", tmp);
					sAppend(tmp1->s, tmp);
					sAppend(")", tmp);
				}
				else
					sAppend(tmp1->s, tmp);

				sAppend(tok, tmp);

				if (hasOperator(tmp2->s))
				{
					sAppend("(", tmp);
					sAppend(tmp2->s, tmp);
					sAppend(")", tmp);
				}
				else
					sAppend(tmp2->s, tmp);

				push(tmp, &stk);
			}
		}
		else
		{
			/* non-number and non-operator encountered, we're done grabbing the input string */
			break;
		}
		tok = strtok(NULL, delim);
	}

	if (!invalid_string) pop(out, &stk);

	stack_free(&stk);
	sFree(tmp);
	free(tmp);
	sFree(tmp1);
	free(tmp1);
	sFree(tmp2);
	free(tmp2);
}


function calc(str_t *in, fact_obj_t *fobj)
{

	/*
	Dijkstra's shunting algorithm 

	While there are tokens to be read: 
		* Read a token. 
		* If the token is a number, then add it to the output queue. 
		* If the token is a function token, then push it onto the stack. 
		* If the token is a function argument separator (e.g., a comma): 
			* Until the topmost element of the stack is a left parenthesis, pop the element onto the 
				output queue. If no left parentheses are encountered, either the separator was misplaced 
				or parentheses were mismatched. 
		* If the token is an operator, o1, then: 
			* while there is an operator, o2, at the top of the stack, and either 
				o1 is associative or left-associative and its precedence is less than 
				(lower precedence) or equal to that of o2, or o1 is right-associative and its precedence 
				is less than (lower precedence) that of o2,
				* pop o2 off the stack, onto the output queue; 
			* push o1 onto the operator stack. 
		* If the token is a left parenthesis, then push it onto the stack. 
		* If the token is a right parenthesis: 
			* Until the token at the top of the stack is a left parenthesis, pop operators off the stack 
				onto the output queue. 
			* Pop the left parenthesis from the stack, but not onto the output queue. 
			* If the token at the top of the stack is a function token, pop it and onto the output queue. 
			* If the stack runs out without finding a left parenthesis, then there are mismatched parentheses. 
	* When there are no more tokens to read: 
		* While there are still operator tokens in the stack: 
			* If the operator token on the top of the stack is a parenthesis, then there are mismatched 
				parenthesis. 
			* Pop the operator onto the output queue. 
	Exit. 

		In addition, we have immediate operators to deal with.	for post immediates (i.e. 4!, 9#) add
	to the output queue, just like a number

	a couple test cases
	trial(4+5^78*(81345-4),21345)

	*/

	//////////////////////////////////////////////
	// as this is primarily a factoring engine frontend, create a factorization 
	// object and pass it around through all the feval calls.	Those functions
	// which do factoring will know what to do with it.	If the expression turns 
	// out to have nothing to do with factorization, then the object will not be
	// used and can just be freed at the end of this routine.
	// ///////////////////////////////////////////

	int i,retval,na,func,j,k;
	bstack_t stk;		//general purpose stack
	str_t *tmp;			//temporary str_t
	str_t *post;		//post fix expression
	char **tokens;		//pointer to an array of strings holding tokens
	char *tok;
	char *strN;			/* use this to make a copy of the post-fix string */
	char delim[2];
	int *token_types;	//type of each token
	int num_tokens;		//number of tokens in the array.
	int varstate;
	mpz_t tmpz;
	
	retval = 0;

	//initialize and find tokens
	token_types = (int *)malloc(100 * sizeof(int));
	tokens = tokenize(in->s, token_types, &num_tokens);
	if (tokens == NULL)
	{		
		free(token_types);
		return 1;
	}

	stack_init(20,&stk,STACK);
	tmp = (str_t *)malloc(sizeof(str_t));
	sInit(tmp);
	for (i=0;i<5;i++)
		mpz_init(operands[i]);
	mpz_init(tmpz);
	post = (str_t *)malloc(sizeof(str_t));
	sInit(post);

	//run the shunting algorithm
	i=0;
	post->s[0] = '\0';
	while (i<num_tokens)
	{
		switch (token_types[i])
		{
		case 3:
			//NUM
			//to num output queue
			sAppend(" ",post);
			sAppend(tokens[i],post);
			break;
		case 6:
			//LP
			toStr(tokens[i],tmp);
			push(tmp,&stk);
			break;
		case 7:
			//string (function or variable name)
			varstate = get_uvar(tokens[i],tmpz);
			if (varstate == 0)
			{
				//found a variable with that name, copy it's value
				//to num output queue
				sAppend(" ",post);
				sAppend(tokens[i],post);
			}
			else if (varstate == 2)
			{
				//do nothing, special case
			}
			else if (getFunc(tokens[i],&na) >= 0) 
			{
				//valid function, push it onto stack
				toStr(tokens[i],tmp);
				push(tmp,&stk);
			}
			else
			{
				//not a variable and not valid num
				printf("unrecognized token: %s\n",tokens[i]);
				retval=1;
				goto free;
			}
			break;
		case 9:
			//comma (function argument separator)
			while (1)
			{
				if (pop(tmp,&stk) == 0)
				{
					//stack empty and we are still looking for a LP
					printf("bad function separator position or mismatched parens\n");
					retval = 1;
					goto free;
				}

				if (strcmp(tmp->s,"(") == 0)
				{
					//found a left paren.	put it back and continue
					push(tmp,&stk);
					break;
				}
				else
				{
					//copy to output operator queue
					sAppend(" ",post);
					sAppend(tmp->s,post);
				}
			}
			break;
		case 5:
			//right paren
			while (1)
			{
				if (pop(tmp,&stk) == 0)
				{
					//stack empty and we are still looking for a LP
					printf("mismatched parens\n");
					retval = 1;
					goto free;
				}

				if (strcmp(tmp->s,"(") == 0)
				{
					//found a left paren.	ignore it.
					if (pop(tmp,&stk) != 0)
					{
						//is the top of stack a function?
						if ((getFunc(tmp->s,&na) >= 0) && (strlen(tmp->s) > 1))
						{
							//the extra check for strlen > 1 fixes
							//the case where the string is an operator, not
							//a function.	for multichar operators this won't work
							//instead should probably separate out the operators
							//from the functions in getFunc


							//yes, put it on the output queue as well
							sAppend(" ",post);
							sAppend(tmp->s,post);
						}
						else
						{
							//no, put it back
							push(tmp,&stk);
						}
					}
					break;
				}
				else
				{
					//copy to output queue
					sAppend(" ",post);
					sAppend(tmp->s,post);
				}
			}
			break;
		case 4:
			//operator
			while (pop(tmp,&stk))
			{
				if (strlen(tmp->s) == 1 && getOP(tmp->s[0]) > 0)
				{
					//its an operator
					//check the precedence
					if (op_precedence(tmp->s,tokens[i],getAssoc(tmp->s)))
					{
						//push to output op queue
						sAppend(" ",post);
						sAppend(tmp->s,post);
					}
					else
					{
						//put the tmp one back and bail
						push(tmp,&stk);
						break;
					}
				}
				else
				{
					//its not an operator, put it back and bail
					push(tmp,&stk);
					break;
				}
			}
			//push the current op onto the stack.
			toStr(tokens[i],tmp);
			push(tmp,&stk);

			break;
		case 2:
			//post unary operator
			//I think we can jush push these into the output operator queue
			toStr(tokens[i],tmp);
			sAppend(" ",post);
			sAppend(tmp->s,post);
			break;
		}
		i++;
	}

	//now pop all operations left on the stack to the output queue
	while (pop(tmp,&stk))
	{
		if (strcmp(tmp->s,"(") == 0 || strcmp(tmp->s,")") == 0)
		{
			printf("mismatched parens\n");
			retval = 1;
			goto free;
		}
		sAppend(" ",post);
		sAppend(tmp->s,post);
	}

	//free the input tokens
	for (i=0;i<num_tokens;i++)
		free(tokens[i]);
	free(tokens);

	//process the output postfix expression:
	//this can be done with a simple stack
	//all tokens are separated by spaces
	//all tokens consist of numbers or functions

	/* try to grab the input number as an expression, if found, store it in fobj->str_N */
	strN = strdup(post->s); /* make a copy of the postfix string */
	if (strN != NULL)
	{
		/* find input expression and store in str_N */
		get_expression(strN, &(fobj->str_N));
		//printf("Found expression: %s\n", fobj->str_N.s);
		free(strN);
	}

	//now evaluate the RPN expression
	//printf("processing postfix expression: %s\n",post->s);
	delim[0] = ' ';
	delim[1] = '\0';
	tok = strtok(post->s,delim);
	if (tok == NULL)
	{
		printf("nothing to process\n");
		goto free;
	}

	do
	{
		//printf("stack contents: ");
		//for (i=0; i<stk.num; i++)
		//	printf("%s ",stk.elements[i]->s);
		//printf("\n");
		//printf("current token: %s\n\n", &tok[0]);

		switch (get_el_type2(tok[0]))
		{
		case NUM:
			//printf("pushing %s onto stack\n",tok);
			toStr(tok,tmp);
			push(tmp,&stk);
			break;
		case AMBIG:
			//could be a number or a function
			//if the next character is a number, this is too
			//else its an operator
			//printf("peeking at tok + 1: %d\n",(int)tok[1]);
			if (get_el_type2(tok[1]) == NUM)
			{
				//printf("pushing %s onto stack\n",tok);
				toStr(tok,tmp);
				push(tmp,&stk);
				break;
			}

			//if not a num, proceed into the next switch (function handle)
		default:
			func = getFunc(tok,&na);
			//printf("processing function %d\n",func);

			if (func >= 0)
			{
				//pop those args and put them in a global array
				for (j=0;j<na;j++)
				{
					//right now we must get all of the operands.
					//somewhere in there we should make allowances
					//for getting a reduced number (i.e. for unary "-"
					//and for variable numbers of arguments
					int r;
					k = pop(tmp,&stk);

					//try to make a number out of it
					//printf("looking at argument %s\n",tmp->s);
					//str2hexz(tmp->s,&tmpz);
					r = mpz_set_str(tmpz, tmp->s, 0);
					//printf("found numerical argument %s\n",z2decstr(&tmpz,&gstr1));

					//tmpz with size zero means this thing isn't a number
					//if (tmpz.size == 0 || k == 0)
					if (r < 0 || k == 0)
					{
						//didn't get the expected number of arguments
						//for this function.	This may be ok, if the
						//function accepts varable argument lists.
						//feval will handle it.

						//put the non-num back
						if (k != 0)
							push(tmp,&stk);

						//and bail
						break;
					}
					else
					{
						//it is a number, put it in the operand pile
						//str2hexz(tmp->s,&operands[na-j-1]);
						mpz_set(operands[na-j-1], tmpz);
					}
				}

				na = j;
				//call the function evaluator with the 
				//operator string and the number of args available
				na = feval(func,na,fobj);

				//put result back on stack
				for (j=0;j<na;j++)
				{
					//z2hexstr(&operands[j],tmp);
					int sz = mpz_sizeinbase(operands[j], 10) + 10;
					if (tmp->alloc < sz)
					{
						tmp->s = (char *)realloc(tmp->s, sz * sizeof(char));
						tmp->alloc = sz;
					}
					mpz_get_str(tmp->s, 10, operands[j]);
					tmp->nchars = strlen(tmp->s)+1;
					push(tmp,&stk);
				}
			}
			else
			{
				int sz;

				get_uvar(tok,tmpz);
				sz = mpz_sizeinbase(tmpz, 10) + 10;
				if (gstr1.alloc < sz)
				{
					gstr1.s = (char *)realloc(gstr1.s, sz * sizeof(char));
					gstr1.alloc = sz;
				}
				mpz_get_str(gstr1.s, 10, tmpz);
				gstr1.nchars = strlen(gstr1.s)+1;
				sCopy(tmp,&gstr1);
				push(tmp,&stk);
			}
			break;
		}

		tok = strtok((char *)0,delim);
	} while (tok != NULL);
	pop(in,&stk);


free:
	for (i=0;i<5;i++)
		mpz_clear(operands[i]);
	free(token_types);
	stack_free(&stk);
	mpz_clear(tmpz);
	sFree(tmp);
	free(tmp);
	sFree(post);
	free(post);
	return retval;

}

